# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Transmission and reflection techniques for state transition 0-1 and 1-2.
In here TR stands for transmission and reflection, and we derive the ideas of doing such method in space
01 and 12 based on Qiskit documents of doing calibration. Users can find it here:
https://qiskit.org/textbook/ch-quantum-hardware/calibrating-qubits-pulse.html
"""
import numpy as np

from qiskit.circuit import Gate, QuantumCircuit, Parameter
from qiskit_ibm_provider.job import job_monitor

from src.backend.backend_ibm import CustomProvider
from src.pulse import Pulse01, Pulse12
from src.constant import QubitParameters
from src.utility import fit_function
from src.calibration.utility import _SetAttribute
from src.pulse_creation import GateSchedule
from src.analyzer import DataAnalysis

from numpy.typing import NDArray
from typing import List, Optional

mhz_unit = QubitParameters.MHZ.value
ghz_unit = QubitParameters.GHZ.value


def set_up_freq(center_freq: float,
                freq_span: int = 40,
                freq_step: float = 0.5) -> NDArray:
    """ Set up frequency range in NDArray form

    Arg:
        default_freq: Either DEFAULT_F01 or DEFAULT_F12
    Returns:
        frequency_range: in GHz
    """
    max_freq = center_freq + freq_span * mhz_unit / 2
    min_freq = center_freq - freq_span * mhz_unit / 2

    freq_ghz = np.arange(min_freq / ghz_unit, max_freq / ghz_unit, freq_step * mhz_unit / ghz_unit)
    return freq_ghz


class _TR(_SetAttribute):
    """ The class act as an abstract class for transmission and reflection technique used in pulse model
    TR typical flow: set up pulse and gates -> submit job to IBM -> Analyze the resulted pulse
    -> return Pulse model and save job_id.
    An example of this flow::
        from qutritium.calibration.transmission_reflection import TR01, TR12
        from qutritium.pulse import Pulse01, Pulse12

        pulse01 = Pulse01(duration=144, x_amp=0.2)
        pulse12 = Pulse12(pulse01=pulse01, duration=pulse01.duration, x_amp=pulse01.x_amp)
        ... (TR_children_classes)

    Note:
        * You should not instantiate _TR as this is am abstract class. Use TR01 and TR12 instead
        * Users are recommended to not initialize pulse model 01 and 12 and let the pulse_model parameter = None
        . This class will instantiate the pulse and return to users for further usage

    Here is list of attributes available on the abstract "_TR" class, excluding SharedAttr attr:
        * frequency: in Hz
        * freq_sweeping_range_ghz: It is freq_sweeping_range but in ghz
        # lambda_list: return lambda_list that we use in this class
        * save_data(y_values): save given signal amplitude
        * run_monitor(package): custom run on IBMQ
        * analyze(): get the frequency of pulse after calibration
    """

    def __init__(self,
                 custom_provider: CustomProvider,
                 backend_name: str,
                 num_shots: int,
                 model_space: str,
                 pulse_connected,
                 pulse_model) -> None:
        """ _TR constructor

         Notes:
            * Frequency here is in Hz
            * freq_sweeping_range_ghz: frequency range in GHz -> convert to GHz later\
            * lambda_list: empirical, modified when running multiple experiments

        Args:
            custom_provider: EffProvider instance
            num_shots: number of shots
            backend_name:

        """
        super().__init__(custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space=model_space,
                         pulse_connected=pulse_connected,
                         pulse_model=pulse_model)
        # Need to be modified by the constructor of children classes
        self.default_frequency: float = 0.
        self.freq_sweeping_range_ghz = None
        self.analyzer: Optional[DataAnalysis] = None
        self.package = []

        # Used for fit_function()
        self._lambda_list = [0, 0, 0, 0]
        self._sweep_gate = Gate("sweep", 1, [])
        self._tr_fit: Optional[NDArray] = None

    @property
    def lambda_list(self) -> List[float]:
        return self._lambda_list

    @lambda_list.setter
    def lambda_list(self, val_list: list) -> None:
        if len(val_list) != 4:
            raise ValueError("Lambda list does not have sufficient elements")
        self._lambda_list = val_list

    @property
    def tr_fit(self) -> NDArray:
        return self._tr_fit

    def analyze(self, job_id: Optional[str]) -> float:
        """ Plots IQ Data given freq_range
        Args:
            job_id: may need else we use the inherent submitted_job_id

        Returns:
            freq: Frequency after calculation
        """
        if job_id is None:
            experiment = self.custom_provider.retrieve_job(self.submitted_job)
        else:
            experiment = self.custom_provider.retrieve_job(job_id)
        self.analyzer = DataAnalysis(experiment)

        # Analyze data
        self.analyzer.retrieve_data(average=True)
        fit_params, self._tr_fit = fit_function(self.freq_sweeping_range_ghz, self.analyzer.IQ_data,
                                                lambda x, c1, q_freq, c2, c3:
                                                (c1 / np.pi) * (c2 / ((x - q_freq) ** 2 + c2 ** 2)) + c3,
                                                self.lambda_list)
        freq = fit_params[1] * QubitParameters.GHZ.value
        return freq

    def draw(self) -> None:
        """ Draw the circuit using qiskit standard library
        Implemented in next updates
        """
        pass

    def run_monitor(self,
                    num_shots: Optional[int] = 0,
                    meas_return: str = 'avg',
                    meas_level: int = 1,
                    **kwargs) -> None:
        """ Custom run execute()
        Args:
            num_shots: modify if needed
            meas_level:
            meas_return:

        Returns:

        """
        self.num_shots = num_shots if num_shots != 0 else self.num_shots
        submitted_job = self.backend.run(self.package,
                                         meas_level=meas_level,
                                         meas_return=meas_return,
                                         shots=self.num_shots,
                                         **kwargs)
        self.submitted_job = submitted_job.job_id()
        print(self.submitted_job)
        job_monitor(submitted_job)


class TR01(_TR):
    """ Used specifically for pulse01 model
    An example of this flow::
        from qutritium.backend.backend_ibm import CustomProvider
        from qutritium.pulse import Pulse01
        from qutritium.calibration.transmission_reflection import TR01

        pulse01 = Pulse01(duration=144, x_amp=0.2)
        custom_provider = CustomProvider()
        tr_01 = TR01(custom_provider=custom_provider, pulse_model=pulse01)
        tr_01.prepare_circuit()
        tr_01.run_monitor()
        tr_01.modify_pulse_model()
        print(tr_01.pulse_model)

    Here is list of attributes available on the ''TR_01'' class:
        + prepare_circuit(): implement abstract run_circuit() from ''_TR''
        * modify_pulse_model(): implement abstract modify_pulse_model() from ''_TR''
    """

    def __init__(self,
                 custom_provider: CustomProvider,
                 backend_name: str = "ibm_brisbane",
                 num_shots: int = 4096) -> None:
        """ TR_01 constructor

        Args:
            custom_provider: EffProvider instance
            num_shots: default 4096 shots
            backend_name: default = 'ibmq_manila'

        Returns:
            * Instance of TR01
        """
        super().__init__(custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         pulse_connected=None,
                         pulse_model=None,
                         model_space='01')
        self.lambda_list = [10, 4.9, 1, -2]
        self.default_frequency = int(self.backend_params['drive_frequency'])
        self.freq_sweeping_range_ghz: NDArray = set_up_freq(center_freq=self.default_frequency)

    def prepare_circuit(self) -> None:
        """ Calibrate single qubit state with custom pulse_model
        Notes:
            * The syntax is contingent but the idea stays the same for every update
        """

        self.pulse_model: Pulse01

        # Sweeping
        frequencies_hz = self.freq_sweeping_range_ghz * ghz_unit
        for freq in frequencies_hz:
            qc_sweep = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            # noinspection DuplicatedCode
            qc_sweep.append(self._sweep_gate, [self.qubit])
            freq_schedule = GateSchedule.freq_gaussian(
                backend=self.backend,
                frequency=freq,
                pulse_model=self.pulse_model,
                qubit=self.qubit
            )
            freq = Parameter('freq')
            qc_sweep.measure(self.qubit, self.cbit)
            qc_sweep.add_calibration(self._sweep_gate, (self.qubit,), freq_schedule, [freq])
            self.package.append(qc_sweep)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """ Only used for debugging + getting result from past job
        Args:
            job_id: string representation of submitted job
        """
        self.pulse_model: Pulse01
        self.pulse_model.frequency = self.analyze(job_id=job_id)


class TR12(_TR):
    """ Used specifically for Pulse12

    An example of this flow::
        from qutritium.calibration.transmission_reflection import TR12
        from qutritium.pulse import Pulse12

        pulse01 = Pulse01(duration=144, x_amp=0.2)
        pulse12 = Pulse12(pulse01=pulse01, duration=pulse01.duration, x_amp=pulse01.x_amp)

        tr_12 = TR_12(pulse_model=pulse12)
        tr_12.prepare_circuit()
        tr_12.run_monitor()
        tr_12.modify_pulse_model()
        print(tr_12.pulse_model)

    Here is the list of attributes available "TR_12: class, excluding parents classes:
        * run_circuit(): implement abstract run_circuit() from _TR
        * modify_pulse_model(): implement abstract modify_pulse_model() from _TR
    """

    def __init__(self,
                 pulse_connected: Pulse01,
                 custom_provider: CustomProvider,
                 backend_name: str = "ibm_brisbane",
                 num_shots: int = 4096) -> None:
        """

        Args:
            custom_provider:
            num_shots: default 4096 shots
            backend_name: default = 'ibmq_manila'

        """
        super().__init__(custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         pulse_connected=pulse_connected,
                         pulse_model=None,
                         model_space='12')
        self.lambda_list = [10, 4.8, 1, -2]
        self.default_frequency = int(self.backend_params['drive_frequency'] + self.backend_params['anharmonicity'])
        self.freq_sweeping_range_ghz: NDArray = set_up_freq(center_freq=self.default_frequency)

    def prepare_circuit(self) -> None:
        """ Some logic as TR1, however we add X gate to resemble state 12
        """

        self.pulse_model: Pulse12
        freq_sweeping_range = self.freq_sweeping_range_ghz * ghz_unit
        # Sweeping
        for freq in freq_sweeping_range:
            qc_sweep = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            qc_sweep.x(self.qubit)
            # noinspection DuplicatedCode
            qc_sweep.append(self._sweep_gate, [self.qubit])
            freq_schedule = GateSchedule.freq_gaussian(
                backend=self.backend,
                frequency=freq,
                pulse_model=self.pulse_model,
                qubit=self.qubit
            )
            freq = Parameter('freq')
            qc_sweep.measure(self.qubit, self.cbit)
            qc_sweep.add_calibration(self._sweep_gate, (self.qubit,), freq_schedule, [freq])
            self.package.append(qc_sweep)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """ Only used for debugging + getting result from past job
        Args:
            job_id: string representation of submitted job

        Returns:

        """
        self.pulse_model: Pulse12
        f12 = self.analyze(job_id=job_id)
        self.pulse_model.frequency = f12



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import copy
from qiskit_ibm_provider import IBMBackend
from typing import List, NamedTuple, DefaultDict, Union, Optional, Any
from collections import namedtuple, defaultdict
from src.pulse import Pulse01, Pulse12
from src.pulse_creation import Shift_phase, Set_frequency, GateSchedule
from src.quantumcircuit.qc_elementary_matrices import u_d, r01, r12
from src.quantumcircuit.instruction_structure import Instruction
from src.quantumcircuit.QC import Qutrit_circuit
from qiskit.pulse.schedule import ScheduleBlock
from numpy.typing import NDArray
import matplotlib.pyplot as plt


class Parameter:
    """
    Static class
    """
    @classmethod
    def get_parameters(cls, U: NDArray) -> NamedTuple:
        """
        Decompose su3 matrices to parameters of selected gates
        Args:
            su3: SU3 Matrix

        Returns: Parameters constructing the gates

        """
        pi = np.pi
        params = namedtuple('params', 'theta1 theta2 theta3 phi1 phi2 phi3 phi4 phi5 phi6')
        if np.round(abs(np.absolute(U[2, 2])), 6) == 1:
            if np.round(abs(np.absolute(U[0, 0])), 6) != 0:
                theta_1 = phi_1 = theta_2 = phi_2 = 0
                phi_4 = np.angle(U[2, 2])
                phi_5 = np.angle(U[1, 1])
                phi_6 = np.angle(U[0, 0])
                # phi_3 = phi_6 - pi/2 - np.angle(U[0, 1])
                phi_3 = np.angle(U[1, 0]) - phi_5 + pi / 2
                theta_3 = 2 * np.arccos(np.round(np.absolute(U[1, 1]), 6))
            else:
                theta_1 = phi_1 = theta_2 = phi_2 = phi_3 = 0
                theta_3 = 2 * np.arccos(np.round(np.absolute(U[1, 1]), 6))
                phi_4 = np.angle(U[2, 2])
                phi_6 = np.angle(U[0, 1]) + phi_3 + pi / 2
                phi_5 = np.angle(U[1, 0]) - phi_3 + pi / 2
        elif np.round(abs(np.absolute(U[2, 2])), 6) == 0:
            theta_1 = 2 * np.arccos(np.round(np.absolute(U[2, 1]), 6))
            theta_2 = pi
            theta_3 = 2 * np.arccos(np.round(np.absolute(U[1, 2]), 6))
            phi_1 = phi_2 = phi_3 = 0
            if np.round(abs(np.absolute(U[2, 0])), 6) != 0:
                phi_4 = np.angle(-U[2, 0])
                if np.round(abs(np.absolute(U[0, 2])), 6) != 0:
                    phi_5 = np.angle(-U[1, 1])
                    phi_6 = np.angle(-U[0, 2])
                else:
                    phi_5 = np.angle(U[1, 2]) + pi / 2
                    phi_6 = np.angle(U[0, 1]) + pi / 2
            if np.round(abs(np.absolute(U[1, 0])), 6) != 0:
                phi_4 = np.angle(U[2, 1]) + pi / 2
                phi_5 = np.angle(U[1, 0]) + pi / 2
                phi_6 = np.angle(-U[0, 2])
            if np.round(abs(np.absolute(U[0, 0])), 6) != 0:
                phi_4 = np.angle(U[2, 1]) + pi / 2
                phi_5 = np.angle(U[1, 2]) + pi / 2
                phi_6 = np.angle(U[1, 1])
        else:
            phi_4 = np.angle(U[2, 2])
            theta_2 = 2 * np.arccos(np.round(np.absolute(U[2, 2]), 6))
            phi_2 = np.angle(U[2, 1]) - phi_4 + pi / 2
            phi_1 = np.angle(-U[2, 0]) - phi_2 - phi_4
            theta_1 = 2 * np.arccos(np.round(np.absolute(U[2, 1]) / np.sin(theta_2 / 2), 6))
            theta_3 = 2 * np.arccos(np.round(np.absolute(U[1, 2]) / np.sin(theta_2 / 2), 6))
            phi_5 = np.angle(U[1, 2]) + phi_2 + pi / 2
            phi_3 = np.angle(np.cos(theta_1 / 2) * np.cos(theta_2 / 2) * np.cos(theta_3 / 2) - U[1, 1] * np.exp(
                -1j * phi_5)) + phi_1
            phi_6 = np.angle(-U[0, 2]) + phi_3 + phi_2
        paras = params(theta_1, theta_2, theta_3, phi_1, phi_2, phi_3, phi_4, phi_5, phi_6)
        return paras


class SU3_matrices:

    def __init__(self, su3: NDArray, qutrit_index: int, n_qutrits: int) -> None:
        """

        Args:
            su3:
            qutrit_index:
            n_qutrits:
        """
        # Check dimensions
        assert su3.shape[0] == 3
        assert su3.shape[1] == 3
        self.su3: NDArray = su3
        self.qutrit_index = qutrit_index
        self.n_qutrits = n_qutrits
        self.parameters: NamedTuple = Parameter.get_parameters(U=self.su3)

    def unitary_diagonal(self) -> NDArray:
        """

        :return:
        """
        return u_d(phi_1=getattr(self.parameters, 'phi6'),
                   phi_2=getattr(self.parameters, 'phi5'),
                   phi_3=getattr(self.parameters, 'phi4'))

    def rotation_theta3_01(self) -> NDArray:
        """

        :return:
        """
        return r01(phi=getattr(self.parameters, 'phi3'), theta=getattr(self.parameters, 'theta3'))

    def rotation_theta1_01(self) -> NDArray:
        """

        :return:
        """
        return r01(getattr(self.parameters, 'phi1'), getattr(self.parameters, 'theta1'))

    def rotation_theta2_12(self) -> NDArray:
        """

        :return:
        """
        return r12(getattr(self.parameters, 'phi2'), getattr(self.parameters, 'theta2'))

    def reconstruct(self) -> NDArray:
        """

        :return:
        """
        return (
                self.unitary_diagonal()
                @ self.rotation_theta3_01()
                @ self.rotation_theta2_12()
                @ self.rotation_theta1_01()
        )

    def native_list(self) -> List[Union[List[float], List[Instruction]]]:
        """

        :return:
        """
        phase01 = float(getattr(self.parameters, 'phi6') - getattr(self.parameters, 'phi5'))
        phase12 = float(getattr(self.parameters, 'phi5') - getattr(self.parameters, 'phi4'))
        return [np.array([phase01, phase12]), [Instruction(gate_type='g01', first_qutrit_set=self.qutrit_index,
                                                           second_qutrit_set=None,
                                                           n_qutrit=self.n_qutrits,
                                                           parameter=[getattr(self.parameters, 'theta1'),
                                                                      getattr(self.parameters, 'phi1')]),
                                               Instruction(gate_type='g12', first_qutrit_set=self.qutrit_index,
                                                           second_qutrit_set=None,
                                                           n_qutrit=self.n_qutrits,
                                                           parameter=[getattr(self.parameters, 'theta2'),
                                                                      getattr(self.parameters, 'phi2')]),
                                               Instruction(gate_type='g01', first_qutrit_set=self.qutrit_index,
                                                           second_qutrit_set=None,
                                                           n_qutrit=self.n_qutrits,
                                                           parameter=[getattr(self.parameters, 'theta3'),
                                                                      getattr(self.parameters,
                                                                              'phi3')])]]

    def decomposed_into_qc(self):
        decomposed_qc = Qutrit_circuit(n_qutrit=self.n_qutrits, initial_state=None)
        decomposed_qc.add_gate("g01", first_qutrit_set=self.qutrit_index, parameter=[getattr(self.parameters, 'theta1'),
                                                                                     getattr(self.parameters, 'phi1')])
        decomposed_qc.add_gate("g12", first_qutrit_set=self.qutrit_index, parameter=[getattr(self.parameters, 'theta2'),
                                                                                     getattr(self.parameters, 'phi2')])
        decomposed_qc.add_gate("g01", first_qutrit_set=self.qutrit_index, parameter=[getattr(self.parameters, 'theta3'),
                                                                                     getattr(self.parameters, 'phi3')])
        decomposed_qc.add_gate("u_d", first_qutrit_set=self.qutrit_index, parameter=[getattr(self.parameters, 'phi6'),
                                                                                     getattr(self.parameters, 'phi5'),
                                                                                     getattr(self.parameters, 'phi4')])
        return decomposed_qc

    def __str__(self) -> str:
        """

        :return:
        """
        return f"U_diagonal:\n{self.unitary_diagonal()}\n" \
               f"R_theta1:\n{self.rotation_theta1_01()}\n" \
               f"R_theta2:\n{self.rotation_theta2_12()}\n" \
               f"R_theta3:\n{self.rotation_theta3_01()}\n"

    def __repr__(self) -> str:
        """

        :return:
        """
        return f"U_diagonal:\n{self.unitary_diagonal()}\n" \
               f"R_theta1:\n{self.rotation_theta1_01()}\n" \
               f"R_theta2:\n{self.rotation_theta2_12()}\n" \
               f"R_theta3:\n{self.rotation_theta3_01()}\n"


class Pulse_Wrapper:
    """

    """

    def __init__(self, pulse01: Pulse01,
                 pulse12: Pulse12,
                 /,
                 qc: Qutrit_circuit,
                 native_gates: Optional[List[str]],
                 backend: IBMBackend) -> None:
        """
        :param qc:
        """
        self.qc = qc
        self.n_qutrit = qc.n_qutrit
        self.ins_list = qc.operation_set
        self._su3_dictionary: DefaultDict[str, Any] = defaultdict()
        self.pulse01 = pulse01
        self.pulse12 = pulse12
        self.backend = backend
        self.native_gates = native_gates if native_gates else ['u_d', 'rx', 'ry', 'rz']
        self.pulse_wrapper = []
        self.qiskit_schedule = None
        self.accumulated_phase = [np.array([0.0, 0.0])]

    def decompose(self) -> None:
        """

        Convert to SU3_matrices for further decomposition

        """
        operation_set = self.ins_list
        n_qutrit = self.qc.n_qutrit

        for instruction in operation_set:
            if type(instruction) == str:
                continue
            if (ins_type := instruction.type()) not in self.native_gates:
                decomposed_res = SU3_matrices(instruction.gate_matrix,
                                              qutrit_index=0, n_qutrits=n_qutrit).native_list()
                self._su3_dictionary[ins_type] = decomposed_res[1]
                self.accumulated_phase.append(self.accumulated_phase[-1] + decomposed_res[0])
            else:
                self._su3_dictionary[ins_type] = [instruction]

    def convert_to_pulse_model(self):
        """
        Convert to internal Pulse Model
        """
        cnt = 0
        alpha = 4.55047476
        beta = 0.22474651
        gamma = 0.03502484
        advance_phase = np.array([0.0, 0.0])  # Implementation of phase advance
        for instruction in self.ins_list:
            if type(instruction) == str:
                continue
            ins_type = instruction.type()
            phase_ud = self.accumulated_phase[cnt]
            for ins in self._su3_dictionary[ins_type]:
                gate_type = ins.type()
                if gate_type[0:2] == "rx":
                    if gate_type[2:4] == "01":
                        pulse = copy.deepcopy(self.pulse01)
                        pulse.x_amp = (pulse.x_amp / np.pi) * ins.parameter[0]
                        phase_operator = Shift_phase(value=phase_ud[0],
                                                     channel=instruction.first_qutrit, subspace="01",
                                                     backend=self.backend)
                    elif gate_type[2:4] == "12":
                        pulse = copy.deepcopy(self.pulse12)
                        pulse.x_amp = (pulse.x_amp / np.pi) * ins.parameter[0]
                        phase_operator = Shift_phase(value=phase_ud[1],
                                                     channel=instruction.first_qutrit, subspace="12",
                                                     backend=self.backend)
                    else:
                        raise Exception("The gate can not be decomposed to pulse")
                elif gate_type[0:2] == "rz":
                    if gate_type[2:4] == "01":
                        pulse = "01"
                        phase_operator = Shift_phase(value=phase_ud[0],
                                                     channel=instruction.first_qutrit, subspace="01",
                                                     backend=self.backend)
                    elif gate_type[2:4] == "12":
                        pulse = "12"
                        phase_operator = Shift_phase(value=phase_ud[1],
                                                     channel=instruction.first_qutrit, subspace="12",
                                                     backend=self.backend)
                    else:
                        raise Exception("The gate can not be decomposed to pulse")
                elif gate_type == 'g01':
                    pulse = copy.deepcopy(self.pulse01)
                    pulse.x_amp = (pulse.x_amp / np.pi) * ins.parameter[0]
                    phase_operator = Shift_phase(value=phase_ud[0] + ins.parameter[1] + advance_phase[0],
                                                 # Adding
                                                 # phase advance
                                                 channel=instruction.first_qutrit, subspace="01",
                                                 backend=self.backend)
                    advance_phase[1] += alpha  # Accumulate phase advance
                elif gate_type == 'g12':
                    pulse = copy.deepcopy(self.pulse12)
                    pulse.x_amp = (pulse.x_amp / np.pi) * ins.parameter[0]
                    phase_operator = Shift_phase(value=phase_ud[1] + ins.parameter[1] + advance_phase[1],
                                                 # Adding
                                                 # phase advance
                                                 channel=instruction.first_qutrit, subspace="12",
                                                 backend=self.backend)
                    advance_phase[0] += beta
                    advance_phase[1] += gamma  # Accumulate phase advance
                else:
                    raise Exception("The gate can not be decomposed to pulse")
                self.pulse_wrapper.append([pulse, phase_operator, instruction.first_qutrit])
            cnt += 1

    def pulse_model_to_qiskit(self):
        """
        Returns: Convert to Qiskit Pulse Model
        """
        schedule = ScheduleBlock()
        for pul in self.pulse_wrapper:
            if type(pul[0]) in [Pulse01, Pulse12]:
                tmp_pulse = GateSchedule.x_amp_gaussian(pulse_model=pul[0], qubit=pul[2],
                                                        backend=self.backend, x_amp=pul[0].x_amp)
                schedule += pul[1].generate_qiskit_phase_offset(gate_pulse=tmp_pulse)
            else:
                if pul[1].subspace == "01":
                    freq_op = Set_frequency(value=self.pulse01.frequency, channel=pul[2], backend=self.backend)
                elif pul[1].subspace == "12":
                    freq_op = Set_frequency(value=self.pulse12.frequency, channel=pul[2], backend=self.backend)
                else:
                    raise Exception("The pulse can not be translated to Qiskit pulse")
                freq_schedule = freq_op.generate_qiskit_freq()
                phase_schedule = pul[1].generate_qiskit_phase(coeff=1.0)
                schedule = schedule + freq_schedule + phase_schedule
        self.qiskit_schedule = schedule
        return schedule

    def print_decompose_ins(self):
        """

        Check the ins after decomposition

        """
        cnt = 0
        for ins in self.ins_list:
            if type(ins) == str:
                continue
            print("Phase accumulated: " + str(self.accumulated_phase[cnt]))
            for i in self._su3_dictionary[ins.type()]:
                i.print()
            cnt += 1

    def print_decompose_pulse(self):
        """

        Check the pulses after decomposition

        """
        for pul in self.pulse_wrapper:
            print(pul)

    def print_qiskit_schedule(self):
        """

        Check the ins after decomposition

        """

        if self.qiskit_schedule is not None:
            print(self.qiskit_schedule)
            self.qiskit_schedule.draw()
            plt.savefig('Qiskit_schedule.png', bbox_inches='tight')
            plt.show()
        else:
            raise Exception("Required conversion to Qiskit ScheduleBlock")

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:
        return ""


# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Support the construction of Qutrit tomography
"""
import copy

import numpy as np
from src.quantumcircuit.QC import Qutrit_circuit

Gell_man_matrices = [
    np.array([[0, 1, 0],
              [1, 0, 0],
              [0, 0, 0]]),
    np.array([[0, -1j, 0],
              [1j, 0, 0],
              [0, 0, 0]]),
    np.array([[1, 0, 0],
              [0, -1, 0],
              [0, 0, 0]]),
    np.array([[0, 0, 1],
              [0, 0, 0],
              [1, 0, 0]]),
    np.array([[0, 0, -1j],
              [0, 0, 0],
              [1j, 0, 0]]),
    np.array([[0, 0, 0],
              [0, 0, 1],
              [0, 1, 0]]),
    np.array([[0, 0, 0],
              [0, 0, -1j],
              [0, 1j, 0]]),
    1 / np.sqrt(3) * np.array([[1, 0, 0],
                               [0, 1, 0],
                               [0, 0, -2]])]


class Tomography:
    """
    The class support the construction of Qutrit tomography given a Qutrit circuit
    """
    def __init__(self, circuit: Qutrit_circuit):
        """
        Args:
            circuit: Qutrit circuit to be used for tomography
        """
        self.circuit = circuit
        self.n_qutrit = circuit.n_qutrit
        self._exp_circuit = [self.circuit]
        if circuit.measurement_flag:
            raise Exception("The given circuit should not contains measurement")
        self.tomo_exp_op = [["rx01", "", [np.pi/2, 0]], ["ry01", "", [np.pi/2, 0]], ["rx01", "", [np.pi, 0]],
                            ["rx12", "", [np.pi/2, 0]], ["ry12", "", [np.pi/2, 0]], ["rx01", "rx12", [np.pi, np.pi/2]],
                            ["rx01", "ry12", [np.pi, np.pi/2]], ["rx01", "rx12", [np.pi, np.pi]]]

    def construct_tomography_exp(self) -> None:
        """
        Returns: Construct the qutrit circuits experiments for tomography
        """
        for i in self.tomo_exp_op:
            tmp_circuit = copy.copy(self.circuit)
            if i[1] == "":
                tmp_circuit.add_gate(gate_type=i[0], first_qutrit_set=0, parameter=i[2][0], to_all=True)
            else:
                tmp_circuit.add_gate(gate_type=i[0], first_qutrit_set=0, parameter=i[2][0], to_all=True)
                tmp_circuit.add_gate(gate_type=i[1], first_qutrit_set=0, parameter=i[2][1], to_all=True)
            tmp_circuit.measure_all()
            self._exp_circuit.append(tmp_circuit)

    def execute_tomography(self):
        return

    def tomography_exp(self):
        return self._exp_circuit



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Fine Tune techniques"""
import numpy as np
from numpy.typing import NDArray

from qiskit import QuantumCircuit
from qiskit.circuit import Gate

from src.calibration.utility import _SharedAttr
from src.backend.backend_ibm import EffProvider
from src.pulse import Pulse01, Pulse12
from src.pulse_creation import GateSchedule

from typing import Union, List, Optional


class _FineTune(_SharedAttr):
    """
    The class act as provider + regulator for Fine Tune techniques
    """

    def __int__(self, pulse_model: Union[Pulse01, Pulse12],
                eff_provider: EffProvider, backend_name: str,
                num_shots: int, alpha_lb: float,
                alpha_ub: float, data_points: int) -> None:
        """

        Args:
            pulse_model:
            eff_provider:
            backend_name:
            num_shots:
            alpha_lb:
            alpha_ub:

        Returns:

        """
        super().__init__(eff_provider=eff_provider, backend_name=backend_name,
                         pulse_model=pulse_model, num_shots=num_shots)
        self.lower_bound = self.pulse_model.x_amp - self.pulse_model.x_amp * alpha_lb
        self.upper_bound = self.pulse_model.x_amp + self.pulse_model.x_amp * alpha_ub
        self.ft_amp_range: NDArray = np.linspace(self.lower_bound, self.upper_bound, data_points)

        self.package: List[QuantumCircuit] = []
        # INTERNAL Attributes
        self._ft_gate = Gate("fine_tune", 1, [])

    def run_monitor(self,
                    num_shots: Optional[int],
                    meas_return: str = 'avg',
                    meas_level: int = 1,
                    **kwargs) -> None:
        """

        Args:
            num_shots:
            meas_return:
            meas_level:
            **kwargs:

        Returns:

        """
        pass

    def analyze(self, job_id: Optional[str] = '') -> float:
        ...

    def draw(self) -> None:
        ...


class FT01(_FineTune):
    """

    """

    def __int__(self, pulse_model: Pulse01,
                eff_provider: EffProvider, backend_name: str,
                num_shots: int = 4096, alpha_lb: float = 0.01,
                alpha_ub: float = 0.01, data_points: int = 100) -> None:
        """

        Args:
            pulse_model:
            eff_provider:
            backend_name:
            num_shots:
            alpha_lb:
            alpha_ub:
            data_points:

        Returns:

        """
        super().__int__(eff_provider=eff_provider, pulse_model=pulse_model,
                        backend_name=backend_name, num_shots=num_shots,
                        alpha_lb=alpha_lb, alpha_ub=alpha_ub,
                        data_points=data_points)

    def prepare_circuit(self) -> None:
        """
        """
        for amp in self.ft_amp_range:
            qc_ft = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            ft_schedule = GateSchedule.x_amp_gaussian(
                backend=self.backend,
                pulse_model=self.pulse_model,
                x_amp=amp,
                qubit=self.qubit
            )
            for i in range(1251):
                qc_ft.append(self._ft_gate, [self.qubit])
            qc_ft.add_calibration(self._ft_gate, [self.qubit], ft_schedule)
            qc_ft.measure(self.qubit, self.cbit)
            self.package.append(qc_ft)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """

        Args:
            job_id:

        """
        self.pulse_model: Pulse01
        self.pulse_model.x_amp = self.analyze(job_id)



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
The Qutrit circuit that contains the qutrit gates and measurement
"""
import numpy as np
from typing import List, Union
from numpy.typing import NDArray
from src.quantumcircuit.qc_utility import print_statevector
from src.quantumcircuit.instruction_structure import Instruction


class Qutrit_circuit:
    """
    This class defines the wrapper for Virtual Machine which can be treated as a Quantum Circuit
    """

    def __init__(self, n_qutrit: int, initial_state: Union[NDArray, None]) -> None:
        """

        Args:
            n_qutrit: Number of qutrit
            initial_state: The initial state of the quantum circuit
        """
        self.n_qutrit = n_qutrit
        self.state = []
        self._measurement_result = None
        self._measurement_flag = False
        self._operation_set: List = []
        self._dimension = 3 ** n_qutrit
        if initial_state is not None:
            if initial_state.shape == (self._dimension, 1):
                self.initial_state = initial_state
                self.state = initial_state
            else:
                raise Exception(
                    "The initial state declared does not have correct dimension. The current shape is " + str(
                        initial_state.shape))
        else:
            self.initial_state = np.array([[0] * self._dimension]).transpose()
            self.initial_state[0][0] = 1
            self.state = self.initial_state

    def add_gate(self, gate_type: str,
                 first_qutrit_set: int,
                 second_qutrit_set: int = None,
                 parameter: List[float] = None,
                 to_all: bool = False,
                 is_dagger: bool = False) -> None:
        """
        Args:
            gate_type: the type of qutrit gate
            first_qutrit_set: the index of first qutrit that the gate has effect on
            second_qutrit_set: the index of second qutrit that the gate has effect on (for multi-qutrit gate)
            parameter: the parameter of the gate
            to_all: apply the gate to all qutrit in the circuit or not?
            is_dagger: complex conjugate of the given gate
        Returns:
        """
        if to_all is True and second_qutrit_set is None:
            for i in range(self.n_qutrit):
                ins = Instruction(gate_type=gate_type, n_qutrit=self.n_qutrit,
                                  first_qutrit_set=i, second_qutrit_set=second_qutrit_set,
                                  parameter=parameter, inverse=is_dagger)
                self.operation_set = [ins]

        else:
            ins = Instruction(gate_type=gate_type, n_qutrit=self.n_qutrit,
                              first_qutrit_set=first_qutrit_set, second_qutrit_set=second_qutrit_set,
                              parameter=parameter, inverse=is_dagger)
            self.operation_set = [ins]

    def add_customized_gate(self, gate_type: str,
                            first_qutrit_set: int,
                            second_qutrit_set: int = None,
                            parameter: List[float] = None,
                            to_all: bool = False,
                            is_dagger: bool = False,
                            custom_matrix: NDArray = None) -> None:
        """
        Args:
            gate_type: the type of qutrit gate
            first_qutrit_set: the index of first qutrit that the gate has effect on
            second_qutrit_set: the index of second qutrit that the gate has effect on (for multi-qutrit gate)
            parameter: the parameter of the gate
            to_all: apply the gate to all qutrit in the circuit or not?
            is_dagger: complex conjugate of the given gate
            custom_matrix: gate_matrix for the customized gate
        Returns:
        """
        if to_all is True and second_qutrit_set is None:
            for i in range(self.n_qutrit):
                ins = Instruction(gate_type=gate_type, n_qutrit=self.n_qutrit,
                                  first_qutrit_set=i, second_qutrit_set=second_qutrit_set,
                                  parameter=parameter, inverse=is_dagger, custom=True, custom_matrix=custom_matrix)
                self.operation_set = [ins]

        else:
            ins = Instruction(gate_type=gate_type, n_qutrit=self.n_qutrit,
                              first_qutrit_set=first_qutrit_set, second_qutrit_set=second_qutrit_set,
                              parameter=parameter, inverse=is_dagger, custom=True, custom_matrix=custom_matrix)
            self.operation_set = [ins]

    def measure_all(self):
        """
        Adding measurement in the qutrit circuit
        """
        if self._measurement_flag is True:
            raise Exception("A measurement has already been added to the circuit.")
        else:
            self._measurement_flag = True
            self.operation_set = ['measurement']

    @property
    def operation_set(self) -> List:
        return self._operation_set

    @operation_set.setter
    def operation_set(self, op: List[Union[Instruction, str]]) -> None:
        self._operation_set.extend(op)

    @property
    def measurement_flag(self) -> bool:
        return self._measurement_flag

    def reset_circuit(self) -> None:
        """
        Delete all the elements of the operation set
        """
        self._operation_set.clear()

    # TODO: Add image circuit
    def draw(self):
        """
        Representation of the quantum circuit
        """
        print("Initial state of the circuit: ")
        print_statevector(self.initial_state, self.n_qutrit)
        print("Set of gate on the circuits: ")
        for i in self._operation_set:
            if type(i) == Instruction:
                i.print()
            else:
                print(i)

    def __add__(self, object2):
        if self.n_qutrit != object2.n_qutrit:
            raise Exception("The two circuit has different number of qubits")
        elif self.measurement_flag is True:
            raise Exception("The first circuit contains measurement which is prohibited")
        else:
            self._operation_set += object2.operation_set
            self._measurement_flag = object2.measurement_flag
            return self



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy as np
from numpy.typing import NDArray

pi = np.pi
state_0 = [[1], [0], [0]]
state_1 = [[0], [1], [0]]
state_2 = [[0], [0], [1]]


def x_plus() -> NDArray:
    return np.array([[0, 0, 1],
                     [1, 0, 0],
                     [0, 1, 0]], dtype=complex)


def x_minus() -> NDArray:
    return np.array([[0, 1, 0],
                     [0, 0, 1],
                     [1, 0, 0]], dtype=complex)


def z01() -> NDArray:
    return np.array([[1, 0, 0],
                     [0, -1, 0],
                     [0, 0, 1]], dtype=complex)


def z12() -> NDArray:
    return np.array([[1, 0, 0],
                     [0, 1, 0],
                     [0, 0, -1]], dtype=complex)


def x01() -> NDArray:
    return np.array([[0, 1, 0],
                     [1, 0, 0],
                     [0, 0, 1]], dtype=complex)


def x12() -> NDArray:
    return np.array([[1, 0, 0],
                     [0, 0, 1],
                     [0, 1, 0]], dtype=complex)


def y01() -> NDArray:
    return np.array([[0, -1j, 0],
                     [1j, 0, 0],
                     [0, 0, 1]], dtype=complex)


def y12() -> NDArray:
    return np.array([[1, 0, 0],
                     [0, 0, -1j],
                     [0, 1j, 0]], dtype=complex)


def rx01(theta: float) -> NDArray:
    return np.array([
        [np.cos(theta / 2), -1j * np.sin(theta / 2), 0],
        [-1j * np.sin(theta / 2), np.cos(theta / 2), 0],
        [0, 0, 1]], dtype=complex)


def rx12(theta: float) -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, np.cos(theta / 2), -1j * np.sin(theta / 2)],
        [0, -1j * np.sin(theta / 2), np.cos(theta / 2)]], dtype=complex)


def rz01(phi: float) -> NDArray:
    return np.array([
        [np.exp(1j * phi), 0, 0],
        [0, 1, 0],
        [0, 0, 1]], dtype=complex)


def rz12(phi: float) -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, np.exp(1j * phi)]], dtype=complex)


def ry01(theta: float) -> NDArray:
    return np.array([
        [np.cos(theta / 2), -np.sin(theta / 2), 0],
        [np.sin(theta / 2), np.cos(theta / 2), 0],
        [0, 0, 1]], dtype=complex)


def ry12(theta: float) -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, np.cos(theta / 2), -np.sin(theta / 2)],
        [0, np.sin(theta / 2), np.cos(theta / 2)]], dtype=complex)


def hdm(omega: float = np.exp(1j * 2 * pi / 3)) -> NDArray:
    return 1 / np.sqrt(3) * np.array([
        [1, 1, 1],
        [1, omega, omega ** 2],
        [1, omega ** 2, omega]], dtype=complex)


def sdg(omega: float = np.exp(1j * 2 * pi / 3)) -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, omega]], dtype=complex)


def tdg(omega: float = np.exp(1j * 2 * pi / 3)) -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, np.power(omega, 1 / 3), 0],
        [0, 0, np.power(omega, -1 / 3)]], dtype=complex)


def Identity() -> NDArray:
    return np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1]], dtype=complex)


def r01(phi: float, theta: float) -> NDArray:
    return rz01(-phi) @ rx01(theta) @ rz01(phi)


def r12(phi: float, theta: float) -> NDArray:
    return rz12(phi) @ rx12(theta) @ rz12(-phi)  # note that R12 is reversed


def u_d(phi_1: float, phi_2: float, phi_3: float) -> NDArray:
    return np.array([[np.exp(1j * phi_1), 0, 0],
                     [0, np.exp(1j * phi_2), 0],
                     [0, 0, np.exp(1j * phi_3)]])


def cnot(first_index: int, second_index: int):
    if second_index == first_index:
        raise Exception("Control qutrit and acting qutrit can not be the same")
    else:
        space = np.abs(first_index - second_index) - 1
        if space == 0:
            spacing = 1
        else:
            spacing = np.eye(3 ** space)
        if second_index < first_index:
            matrix = np.kron(np.kron(state_0 @ np.transpose(state_0), spacing), np.eye(3)) + \
                     np.kron(np.kron(state_1 @ np.transpose(state_1), spacing), x01() @ x12()) + \
                     np.kron(np.kron(state_2 @ np.transpose(state_2), spacing), x12() @ x01())
        else:
            matrix = np.kron(np.kron(np.eye(3), spacing), state_0 @ np.transpose(state_0)) + \
                     np.kron(np.kron(x01() @ x12(), spacing), state_1 @ np.transpose(state_1)) + \
                     np.kron(np.kron(x12() @ x01(), spacing), state_2 @ np.transpose(state_2))
        return np.array(matrix, dtype=complex)



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""This abstract class is meant to refactor attributes appearing in calibration classes"""
from abc import ABC

from typing import Optional

from qiskit import pulse

from src.backend.backend_ibm import CustomProvider
from src.pulse import Pulse, Pulse01, Pulse12
from src.constant import QubitParameters


class _SetAttribute(ABC):
    """
    Notes:
        * This is an abstract class and should not be instantiated as it is only meant for programming convenience
    Here is a list of set attributes used in calibration functions:
        * pulse_model: Either Pulse01 or Pulse12
        * eff_provider: CustomProvider instance
        * backend: IBMBackend which retrieves from eff_provider.backend()
        * backend_params: backend properties with the following format:
            ===============   ===============   =============
            effective_qubit   drive_frequency   anharmonicity
            int               float             float
            ===============   ===============   =============
        * package: initialize package List -> deliver circuits
        * qubit: effective qubit from eff_provider
        * cbit: classical qubit
        * num_shots: for execute function
        * submitted_job: job id in string format
    """

    def __init__(self,
                 custom_provider: CustomProvider,
                 backend_name: str,
                 num_shots: int,
                 model_space: str,
                 pulse_model: Optional[Pulse],
                 pulse_connected=None,
                 drive_duration_ns: float = 120.) -> None:
        """

        Args:
            model_space:
            custom_provider:
            backend_name:
            num_shots:
            pulse_model:
            pulse_connected: User can access the related pulse via Pulse getter, and it is recommended to do so
            drive_duration_ns:
        """
        self.custom_provider: CustomProvider = custom_provider
        self.backend, self.backend_params = self.custom_provider.retrieve_backend_info(backend_name)
        self.model_space = model_space
        self.drive_duration_sec = drive_duration_ns * QubitParameters.ns.value
        self.pulse_connected = pulse_connected
        self.pulse_model = self.set_pulse_model() if not pulse_model else pulse_model

        self.qubit: int = self.backend_params['effective_qubit']
        self.cbit: int = QubitParameters.CBIT.value
        self.num_shots: int = num_shots
        self.submitted_job: str = ''

    def set_pulse_model(self) -> Pulse:
        """
        Set pulse model based ok Qiskit pulse -> to run on IBM devices. A number of functions are demonstrated in Qiskit
        tutorial that users can find here: https://qiskit.org/textbook/ch-quantum-hardware/calibrating-qubits-pulse.html
        """
        granularity = self.backend.configuration().timing_constraints['granularity']

        def get_closest_multiple(value, base_number):
            return int(value + base_number / 2) - (int(value + base_number / 2) % base_number)

        def get_close_multiple_of_granularity(num):
            return get_closest_multiple(num, granularity)

        if self.model_space == '01':
            return Pulse01(x_amp=0.1, drive_duration=get_close_multiple_of_granularity(
                pulse.seconds_to_samples(self.drive_duration_sec)
            ), pulse12=self.pulse_connected)
        elif self.model_space == '12':
            return Pulse12(x_amp=0.1, drive_duration=get_close_multiple_of_granularity(
                pulse.seconds_to_samples(self.drive_duration_sec)
            ), pulse01=self.pulse_connected)
        else:
            raise ValueError("Invalid quantum space!. Only write either: '01' or '12'")

    def prepare_circuit(self) -> None:
        pass

    def modify_pulse_model(self, job_id: str = None) -> None:
        pass



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Backend of the VM that can be used to simulate the Quantum Circuit
"""
import numpy as np
import matplotlib.pyplot as plt
from typing import Dict, List
from numpy.typing import NDArray
from src.quantumcircuit.QC import Qutrit_circuit
from src.quantumcircuit.qc_utility import statevector_to_state
from src.quantumcircuit.instruction_structure import Instruction


class QASM_Simulator:
    """
    The class is used to represent a vm_backend simulator in VM,
    A Quantum Circuit is given as input to the vm_backend and the final result is returned.
    """

    def __init__(self, qc: Qutrit_circuit) -> None:
        """

        :param qc:
        """
        self.circuit = qc
        self.n_qutrit = qc.n_qutrit
        self.initial_state = qc.initial_state
        self.state = self.initial_state
        self._measurement_flag = qc.measurement_flag
        self._operation_set = qc.operation_set
        self._SPAM_error = None
        self._error_meas: List = []
        self._measurement_result: List = []
        self._simulation_flag = False

    def add_SPAM_noise(self, p_prep: float, p_meas: float, error_type: str = 'Pauli_error') -> None:
        """

        Args:
            p_prep: Probability of preparation error
            p_meas: Probability of measurement error
            error_type: The type of error

        Adding State Preparation and Measurement noise to the backend

        """
        if error_type == 'Pauli_error':
            self._error_meas = [('x+', p_meas / 2), ('x-', p_meas / 2), ('I', 1 - p_meas)]
            error_prep = [('x+', p_prep / 2), ('x-', p_prep / 2), ('I', 1 - p_prep)]
            """
            Adding preparation error
            """
            probs_prep = [error_prep[i][1] for i in range(len(error_prep))]
            for i in range(self.n_qutrit):
                choice = np.random.choice(range(len(error_prep)), p=probs_prep)
                error_effect = Instruction(gate_type=error_prep[choice][0], n_qutrit=self.n_qutrit, first_qutrit_set=i,
                                           second_qutrit_set=None, parameter=None)
                self._operation_set.insert(__index=0, __object=error_effect)
            """
            Adding measurement error
            """
            self._SPAM_error = True
        else:
            pass

    def _simulation(self) -> None:
        """

        The simulation process of the vm_backend

        """

        if self._measurement_flag:
            for i in range(len(self._operation_set)-1):
                self.state = np.einsum('ij,jk', self._operation_set[i].effect_matrix, self.state)
        else:
            for i in self._operation_set:
                self.state = np.einsum('ij,jk', i.effect_matrix, self.state)
        self._simulation_flag = True

    def run(self, num_shots: int = 1024) -> None:
        """

        Args:
            num_shots: Number of shots

        Performs measurement through the defined amount of shots.

        """

        if self._simulation_flag is False:
            self._simulation()
        if self._measurement_flag:
            if self._SPAM_error:
                probs_prep = [self._error_meas[i][1] for i in range(len(self._error_meas))]
                for i in range(self.n_qutrit):
                    choice = np.random.choice(range(len(self._error_meas)), p=probs_prep)
                    error_effect = Instruction(gate_type=self._error_meas[choice][0], n_qutrit=self.n_qutrit,
                                               first_qutrit_set=i,
                                               second_qutrit_set=None, parameter=None)
                    self._operation_set.insert(__index=0, __object=error_effect)
            state_coeff, state_construction = statevector_to_state(self.state, self.n_qutrit)
            probs = [np.abs(i) ** 2 for i in state_coeff]
            for i in range(num_shots):
                measure = np.random.choice(range(len(state_construction)), p=probs)
                self._measurement_result.append(state_construction[measure])
        else:
            raise Exception("Your circuit does not contains measurement.")

    def get_counts(self) -> Dict:
        """

        Returns: count of each state

        """
        if self._measurement_result is not None:
            return dict((x, self._measurement_result.count(x)) for x in set(self._measurement_result))
        else:
            raise Exception("You have not made measurement yet.")

    def return_final_state(self) -> NDArray:
        """

        Returns: Final state of the quantum circuit

        """
        if self._simulation_flag is False:
            self._simulation()
        return self.state

    def result(self) -> List:
        """

        Returns: Measurement result

        """
        if self._measurement_result is not None:
            return self._measurement_result
        else:
            raise Exception("You have not made measurement yet.")

    def density_matrix(self) -> NDArray:
        """

        Returns: Density matrix of the current state of the quantum circuit

        """
        if self._simulation_flag is False:
            self._simulation()
        return self.state @ np.transpose(self.state)

    def plot(self, plot_type: str) -> None:
        """

        Args:
            plot_type: Type of plotting

        Returns: Draw graph

        """
        result_dict = self.get_counts()
        if plot_type == "histogram":
            plt.bar(result_dict.keys(), result_dict.values())
        elif plot_type == "line":
            plt.plot(result_dict.keys(), result_dict.values())
        elif plot_type == "dot":
            plt.scatter(result_dict.keys(), result_dict.values())
        else:
            raise Exception("Invalid plotting type!")
        plt.show()



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Unitary Gates: Elementary matrices
"""
from __future__ import annotations
from numpy import ndarray
from numpy.linalg import inv, LinAlgError
from src.quantumcircuit.qc_elementary_matrices import *
from typing import List
import numpy as np

pi = np.pi
state_0 = [[1], [0], [0]]
state_1 = [[0], [1], [0]]
state_2 = [[0], [0], [1]]


def single_matrix_form(gate_type: str, parameter: List[float] = None, omega=np.exp(1j * 2 * pi / 3)):
    """

    Args:
        gate_type: quantum gate type as define in gate_set
        parameter: parameter of rotation gate (if needed)
        omega: omega for phase gate
    Returns: matrix form of the qutrit gate

    """
    if gate_type == 'x01':
        return np.array([[0, 1, 0],
                         [1, 0, 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'rx01':
        return np.array([[np.cos(parameter[0] / 2), -1j * np.sin(parameter[0] / 2), 0],
                         [-1j * np.sin(parameter[0] / 2), np.cos(parameter[0] / 2), 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'g01':
        return np.array([[np.cos(parameter[0] / 2), -1j * np.sin(parameter[0] / 2) * np.exp(-1j * parameter[1]), 0],
                         [-1j * np.sin(parameter[0] / 2) * np.exp(1j * parameter[1]), np.cos(parameter[0] / 2), 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'x12':
        return np.array([[1, 0, 0],
                         [0, 0, 1],
                         [0, 1, 0]], dtype=complex)
    elif gate_type == 'rx12':
        return np.array([[1, 0, 0],
                         [0, np.cos(parameter[0] / 2), -1j * np.sin(parameter[0] / 2)],
                         [0, -1j * np.sin(parameter[0] / 2), np.cos(parameter[0] / 2)]], dtype=complex)
    elif gate_type == 'g12':
        return np.array([[1, 0, 0],
                         [0, np.cos(parameter[0] / 2), -1j * np.sin(parameter[0] / 2) * np.exp(-1j * parameter[1])],
                         [0, -1j * np.sin(parameter[0] / 2) * np.exp(1j * parameter[1]), np.cos(parameter[0] / 2)]],
                        dtype=complex)
    elif gate_type == 'Identity':
        return np.array([[1, 0, 0],
                         [0, 1, 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'x_plus':
        return np.array([[0, 0, 1],
                         [1, 0, 0],
                         [0, 1, 0]], dtype=complex)
    elif gate_type == 'x_minus':
        return np.array([[0, 1, 0],
                         [0, 0, 1],
                         [1, 0, 0]], dtype=complex)
    elif gate_type == 'z01':
        return np.array([[1, 0, 0],
                         [0, -1, 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'rz01':
        return np.array([[np.exp(-1j * parameter[0] / 2), 0, 0],
                         [0, np.exp(1j * parameter[0] / 2), 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'z12':
        return np.array([[1, 0, 0],
                         [0, 1, 0],
                         [0, 0, -1]], dtype=complex)
    elif gate_type == 'rz12':
        return np.array([[1, 0, 0],
                         [0, np.exp(-1j * parameter[0] / 2), 0],
                         [0, 0, np.exp(1j * parameter[0] / 2)]], dtype=complex)
    elif gate_type == 'y01':
        return np.array([[0, -1j, 0],
                         [1j, 0, 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'ry01':
        return np.array([[np.cos(parameter[0] / 2), -np.sin(parameter[0] / 2), 0],
                         [np.sin(parameter[0] / 2), np.cos(parameter[0] / 2), 0],
                         [0, 0, 1]], dtype=complex)
    elif gate_type == 'y12':
        return np.array([[1, 0, 0],
                         [0, 0, -1j],
                         [0, 1j, 0]], dtype=complex)
    elif gate_type == 'ry12':
        return np.array([[1, 0, 0],
                         [0, np.cos(parameter[0] / 2), -np.sin(parameter[0] / 2)],
                         [0, np.sin(parameter[0] / 2), np.cos(parameter[0] / 2)]], dtype=complex)
    elif gate_type == 'hdm':
        return (1 / np.sqrt(3)) * np.array([[1, 1, 1],
                                            [1, omega, np.conj(omega)],
                                            [1, np.conj(omega), omega]], dtype=complex)
    elif gate_type == 'sdg':
        return np.array([[1, 0, 0],
                         [0, 1, 0],
                         [0, 0, omega]], dtype=complex)
    elif gate_type == 'tdg':
        return np.array([[1, 0, 0],
                         [0, np.power(omega, 1 / 3), 0],
                         [0, 0, np.power(omega, -1 / 3)]], dtype=complex)
    elif gate_type == 'u_d':
        return np.array([[np.exp(1j*parameter[0]), 0, 0],
                         [0, np.exp(1j*parameter[1]), 0],
                         [0, 0, np.exp(1j*parameter[2])]], dtype=complex)
    elif gate_type == 'u_ft':
        return (1 / np.sqrt(3)) * np.array([[omega, 1, np.conj(omega)],
                                            [1, 1, 1],
                                            [np.conj(omega), 1, omega]], dtype=complex)
    else:
        raise Exception("This gate is not implemented yet.")


def multi_matrix_form(gate_type: str, first_index: int, second_index: int):
    """

    Args:
        gate_type: quantum gate type as define in gate_set
        first_index: acting qubit
        second_index: control qubit

    Returns: matrix form of the control-qutrit gate

    """

    if gate_type == 'CNOT':
        if second_index == first_index:
            raise Exception("Control qutrit and acting qutrit can not be the same")
        else:
            space = np.abs(first_index - second_index) - 1
            if space == 0:
                spacing = 1
            else:
                spacing = np.eye(3 ** space)
            if second_index < first_index:
                matrix = np.kron(np.kron(state_0 @ np.transpose(state_0), spacing), np.eye(3)) + \
                         np.kron(np.kron(state_1 @ np.transpose(state_1), spacing),
                                 single_matrix_form('x01') @ single_matrix_form('x12')) + \
                         np.kron(np.kron(state_2 @ np.transpose(state_2), spacing),
                                 single_matrix_form('x12') @ single_matrix_form('x01'))
            else:
                matrix = np.kron(np.kron(np.eye(3), spacing), state_0 @ np.transpose(state_0)) + \
                         np.kron(np.kron(single_matrix_form('x01') @ single_matrix_form('x12'), spacing),
                                 state_1 @ np.transpose(state_1)) + \
                         np.kron(np.kron(single_matrix_form('x12') @ single_matrix_form('x01'), spacing),
                                 state_2 @ np.transpose(state_2))
            return np.array(matrix, dtype=complex)


def statevector_to_state(state: np.array, n_qutrit: int):
    """

    Args:
        state: State vector of qutrits
        n_qutrit: number of qutrit

    Returns: state coefficient and state in ket form

    """

    if state.shape != (3 ** n_qutrit, 1):
        raise Exception("The dimension of the state is not align with given qubits")
    state_basis = []
    state_coeff = []
    for i in range(3 ** n_qutrit):
        if abs(complex(state[i][0])) != 0.0:
            state_basis.append(i)
            state_coeff.append(state[i][0])
    state_construction = []
    for j in range(len(state_basis)):
        sta = ''
        tmp = state_basis[j]
        for k in range(n_qutrit):
            sta += str(int(tmp % 3))
            tmp = tmp / 3
        state_construction.append(sta)
    return state_coeff, state_construction


def print_statevector(state: np.array, n_qutrit: int):
    """

    Args:
        state: State vector of qutrits
        n_qutrit: number of qutrit

    Print the state vector of qutrits to ket form
    """
    state_coeff, state_cons = statevector_to_state(state, n_qutrit)
    print("State: ")
    for i in range(len(state_cons)):
        if i == len(state_cons) - 1:
            print(str(state_coeff[i]) + " |" + str(state_cons[i]) + ">")
        else:
            print(str(state_coeff[i]) + " |" + str(state_cons[i]) + "> + ")


def checking_unitary(u: ndarray):
    """

    Args:
        u: Matrix to check

    Returns: True if matrix is unitary, False otherwise

    """
    try:
        p = u @ inv(u)
    except LinAlgError:
        print("The matrix can not inverse")
        return False
    if np.absolute(np.sum(p - np.eye(3))) < 1e-5:
        return True
    else:
        return False



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""List all the available backends and assign their empirically effective qubit"""
from qiskit_ibm_provider import IBMProvider, IBMBackend

from src.constant import  QubitParameters

from typing import DefaultDict, Tuple, Optional

from collections import defaultdict


def set_qubit_dict() -> DefaultDict:
    """ Set initiated qubit for each ibm quantum computers
    Returns:
        Dictionary that has the following format
        ===============  ===============
        IBMBackend Name  Effective Qubit
        'ibm_brisbane'    0
        ===============  ===============
    """
    return defaultdict(lambda: 0)


class CustomProvider(IBMProvider):
    """ Customize IBMProvider
    An example of this flow::

        from qutritium.backend.backend_ibm import EffProvider

        eff_provider = EffProvider()
        backend, _ = eff_provider.backend('ibmq_lima')

        eff_provider.show()

    Notes:
        * This class inherits IBMProvider so the docs is analogous to IBMProvider.

    Here is a list of available attributes in class "EffProvider" class, besides IBMProvider attrs:
        * eff_dict: a dictionary that contains effective qubit number of respective quantum computer
        * backend(): return the wanted backend and its parameters
        * show(): print all available backends
    """

    def __init__(
            self,
            token: Optional[str] = None,
            url: Optional[str] = None,
            name: Optional[str] = None,
            instance: Optional[str] = None,
            proxies: Optional[dict] = None,
            verify: Optional[bool] = None,
            custom_dict: Optional[DefaultDict] = None,
    ) -> None:
        """ Refer to the IBMProvider doc
        Args:
            token:
            url:
            name:
            instance:
            proxies:
            verify:
            custom_dict: Dictionary contains name of the backend and their effective qubits. For example:
                        dict_eff = {'ibmq_lima': 2,
                                    'ibm_nairobi': 6,}

        Returns:
            An instance of EffBackends
        """
        super().__init__(token, url, name, instance, proxies, verify)
        if not self.active_account():
            raise ValueError('Can not find account saved on disk. Please provide token via constructor'
                             ', or save_account() function')
        self.custom_dict: DefaultDict[str, int] = custom_dict if custom_dict else set_qubit_dict()

    def retrieve_backend_info(self, name: str) -> Tuple[IBMBackend, DefaultDict]:
        """

        Args:
            name: name of the quantum computer which appears on API

        Returns:
            Tuple: IBMBackend and its parameters

        """
        backend = self.backends(name=name)[0]
        default_freq = backend.defaults().qubit_freq_est[self.custom_dict[name]]
        anharmonicity = backend.properties().qubits[self.custom_dict[name]][3].value * QubitParameters.GHZ.value
        backend_params = defaultdict(lambda: 0,
                                     {'effective_qubit': self.custom_dict[name],
                                      'drive_frequency': default_freq,
                                      'anharmonicity': anharmonicity})
        # from src.simple_backend_log import write_log
        #
        # write_log(backend)
        return backend, backend_params

    def show(self) -> None:
        """
        Show all available backends based on given provider
        """
        print(f"{'Backend name:':<30}{'# Qubit used:':<40}")
        for available_backend in self.backends():
            name = available_backend.name
            print(f"{name:<30}{self.custom_dict[name]:<40}")

    def qubit_optimization(self):
        ...

    def customize_qubit(self):
        ...



"""Exceptions for pulse model"""


class MissingFrequencyPulse(Exception):
    """Raise if the pulse does not have frequency"""
    pass


class MissingDurationPulse(Exception):
    """Raise if the pulse does not have duration"""
    pass


class MissingAmplitudePulse(Exception):
    """Raise if the pulse does not have x_amp"""
    pass



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Drag Coefficient circuit"""
import numpy as np

from qiskit.circuit import Gate, QuantumCircuit

# from src.analyzer import DataAnalysis
from src.pulse import Pulse01, Pulse12
from src.backend.backend_ibm import CustomProvider
from src.pulse_creation import GateSchedule
from src.calibration.utility import _SetAttribute
from src.calibration.discriminator import DiscriminatorQutrit
from src.exceptions.pulse_exception import (
    MissingDurationPulse,
    MissingFrequencyPulse,
    MissingAmplitudePulse
)

from typing import Union, Optional


class _DRAG(_SetAttribute):
    """
    The class act as provider + regulator for Drag Leakage.
    Drag Leakage flow: set up -> create circuit -> submit job to IBM -> get the result
    """

    def __init__(self,
                 pulse_model: Union[Pulse01, Pulse12],
                 custom_provider: CustomProvider,
                 discriminator_package: DiscriminatorQutrit,
                 model_space: str,
                 pulse_connected: Union[Pulse01, Pulse12],
                 backend_name: str,
                 num_shots: int) -> None:
        """

        Args:
            pulse_model:
            custom_provider:
            discriminator_package:
            backend_name:
            num_shots:

        Raises:
            MissingDurationPulse
            MissingFrequencyPulse
            MissingAmplitudePulse
        """

        if pulse_model.drive_duration == 0:
            raise MissingDurationPulse
        if pulse_model.frequency == 0:
            raise MissingFrequencyPulse
        if pulse_model.x_amp == 0:
            raise MissingAmplitudePulse

        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space=model_space,
                         pulse_connected=pulse_connected)

        self.drag_sweeping_range = np.linspace(-5, 5, 100)
        self.package = discriminator_package.package

        # INTERNAL DESIGN ONLY
        self._drag_gate = Gate('DRAG', 1, [])
        self._delay_gate = Gate('5mus_delay', 1, [])

    def run_monitor(self,
                    num_shots: Optional[int] = 0,
                    meas_return: str = 'single',
                    meas_level: int = 1,
                    **kwargs):
        """

        Args:
            num_shots:
            meas_return:
            meas_level:
            **kwargs:

        """
        from qiskit.tools.monitor import job_monitor
        from qiskit import execute
        from qiskit_ibm_provider import IBMJob

        self.num_shots = num_shots if num_shots != 0 else self.num_shots
        submitted_job: IBMJob = execute(experiments=self.package,
                                        backend=self.backend,
                                        shots=self.num_shots,
                                        meas_level=meas_level,
                                        meas_return=meas_return,
                                        **kwargs)
        self.submitted_job = submitted_job.job_id()
        print(self.submitted_job)
        job_monitor(submitted_job)

    def analyze(self, job_id: Optional[str]) -> float:
        ...

    def draw(self) -> None:
        """

        """
        ...


# noinspection DuplicatedCode
class DRAG01(_DRAG):

    """

    """
    def __init__(self,
                 pulse_model: Pulse01,
                 custom_provider: CustomProvider,
                 discriminator_package: DiscriminatorQutrit,
                 backend_name: str = 'ibm_brisbane',
                 num_shots: int = 4096) -> None:
        """

        Args:
            pulse_model:
            custom_provider:
            discriminator_package:
            backend_name:
            num_shots:
        """
        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         discriminator_package=discriminator_package,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space='01',
                         pulse_connected=pulse_model.pulse12)

    def prepare_circuit(self) -> None:
        """

        """
        for beta in self.drag_sweeping_range:
            beta_circ = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            beta_circ.append(self._delay_gate, [self.qubit])
            beta_sweep_schedule = GateSchedule.drag(
                backend=self.backend,
                pulse_model=self.pulse_model,
                beta=beta,
                qubit=self.qubit
            )
            beta_circ.append(self._drag_gate, [self.qubit])
            beta_circ.add_calibration(self._drag_gate, [self.qubit], beta_sweep_schedule)
            beta_circ.add_calibration(self._delay_gate, [self.qubit], GateSchedule.delay(
                backend=self.backend,
                qubit=self.qubit,
            ))
            beta_circ.measure(self.qubit, self.cbit)
            self.package.append(beta_circ)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """

        Args:
            job_id:

        Returns:

        """
        self.pulse_model: Pulse01
        # Add Drag coefficient
        drag = self.analyze(job_id)
        self.pulse_model.drag_coeff = drag


# noinspection DuplicatedCode
class DRAG12(_DRAG):
    """

    """
    def __init__(self,
                 pulse_model: Pulse12,
                 custom_provider: CustomProvider,
                 discriminator_package: DiscriminatorQutrit,
                 backend_name: str = 'ibm_brisbane',
                 num_shots: int = 4096) -> None:

        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         discriminator_package=discriminator_package,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space='12',
                         pulse_connected=pulse_model.pulse01)

    def prepare_circuit(self) -> None:
        """

        """
        # noinspection DuplicatedCode
        for beta in self.drag_sweeping_range:
            beta_circ = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            beta_circ.append(self._delay_gate, [self.qubit])
            beta_sweep_schedule = GateSchedule.drag(
                backend=self.backend,
                qubit=self.qubit,
                beta=beta,
                pulse_model=self.pulse_model
            )
            beta_circ.append(self._drag_gate, [self.qubit])
            beta_circ.add_calibration(self._drag_gate, [self.qubit], beta_sweep_schedule)
            beta_circ.add_calibration(self._delay_gate, [self.qubit], GateSchedule.delay(
                backend=self.backend,
                qubit=self.qubit,
            ))
            beta_circ.measure(self.qubit, self.cbit)
            self.package.append(beta_circ)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """

        Args:
            job_id:

        Returns:

        """
        self.pulse_model: Pulse12
        # Add Drag coefficient
        drag = self.analyze(job_id)
        self.pulse_model.drag_coeff = drag



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Instruction class that can be called from VM
Used to represent a gate in quantum circuit
"""
from __future__ import annotations
import numpy as np
from typing import Union, Any, List
from numpy.typing import NDArray
# from .qc_elementary_matrices import *
from src.quantumcircuit.qc_utility import multi_matrix_form, single_matrix_form

gate_set: list[Union[str, Any]] = ['Identity',
                                   'x_plus',
                                   'x_minus',
                                   'sdg',
                                   'tdg',
                                   'CNOT',
                                   'g01',
                                   'g12',
                                   'x01',
                                   'x12',
                                   'y01',
                                   'y12',
                                   'z01',
                                   'z12',
                                   'rx01',
                                   'rx12',
                                   'ry01',
                                   'ry12',
                                   'rz01',
                                   'rz12',
                                   'u_d',
                                   'hdm',
                                   'u_ft']


class Instruction:
    """
    The class is used to represent a gate in VM,
    Each gate can be considered as an instruction and each has effect on the final state
    """

    def __init__(self, gate_type: str,
                 n_qutrit: int, first_qutrit_set: int,
                 second_qutrit_set: int | None = 0,
                 parameter: List[float] = None,
                 inverse: bool = False,
                 custom: bool = False,
                 custom_matrix: NDArray = None) -> None:
        self._type = gate_type
        if custom is False:
            self._verify_gate()
        self.n_qutrit: int = n_qutrit
        self.qutrit_dimension: int = 3 ** self.n_qutrit
        self.parameter: List[float] = parameter
        self.first_qutrit: int = first_qutrit_set
        self.second_qutrit: int = second_qutrit_set
        self._is_two_qutrit_gate: bool = False
        self._is_inverse = inverse
        self._is_custom = custom
        if first_qutrit_set > (self.n_qutrit - 1):
            raise Exception("Acting qutrit is not defined")
        if second_qutrit_set is not None:
            self._is_two_qutrit_gate = True
            # if self._type in parameterless_set: eval(self._type)
            if self._is_inverse is False:
                self.gate_matrix = multi_matrix_form(gate_type=self._type, first_index=self.first_qutrit,
                                                     second_index=self.second_qutrit)
            else:
                self.gate_matrix = np.matrix(multi_matrix_form(gate_type=self._type, first_index=self.first_qutrit,
                                                               second_index=self.second_qutrit)).getH()
        else:
            self._is_two_qutrit_gate = False
            # Check if the gate is custom
            if self._is_custom is False:
                self.gate_matrix = single_matrix_form(gate_type=self._type, parameter=self.parameter)
            else:
                self.gate_matrix = custom_matrix
            # Check if we want the inverse of this gate
            if self._is_inverse is True:
                self.gate_matrix = np.matrix(self.gate_matrix).getH()
        self._effect_matrix = self._effect()

    def _effect(self) -> NDArray:
        """

        Returns: the matrix form effect of gate on the quantum state

        """
        if not self._is_two_qutrit_gate:
            if self.n_qutrit == 1:
                return self.gate_matrix
            else:
                if self.first_qutrit == 0:
                    effect_matrix = np.einsum('ik,jl', self.gate_matrix,
                                              np.eye(int(self.qutrit_dimension / 3))).reshape(self.qutrit_dimension,
                                                                                              self.qutrit_dimension)
                else:
                    effect_matrix = np.einsum('ik,jl', np.eye(3 ** self.first_qutrit),
                                              self.gate_matrix).reshape(3 ** (self.first_qutrit + 1),
                                                                        3 ** (self.first_qutrit + 1))
                    effect_matrix = np.einsum('ik,jl', effect_matrix,
                                              np.eye(3 ** (self.n_qutrit - self.first_qutrit - 1))).reshape(
                        self.qutrit_dimension,
                        self.qutrit_dimension)
                return effect_matrix
        else:
            left = min((self.first_qutrit, self.second_qutrit))
            right = max((self.first_qutrit, self.second_qutrit))
            if left == 0:
                effect_matrix = np.einsum('ik,jl', self.gate_matrix,
                                          np.eye(3 ** (self.n_qutrit - right - 1))).reshape(self.qutrit_dimension,
                                                                                            self.qutrit_dimension)
            else:
                effect_matrix = np.einsum('ik,jl', np.eye(3 ** left),
                                          self.gate_matrix).reshape(3 ** (self.first_qutrit + 1),
                                                                    3 ** (self.first_qutrit + 1))
                effect_matrix = np.einsum('ik,jl', effect_matrix,
                                          np.eye(3 ** (self.n_qutrit - right - 1))).reshape(
                    self.qutrit_dimension,
                    self.qutrit_dimension)
            return effect_matrix

    def _verify_gate(self) -> None:
        if self._type not in gate_set:
            raise Exception("This gate is not defined in set of gates")

    @property
    def effect_matrix(self) -> NDArray:
        return self._effect_matrix

    def type(self) -> str:
        return self._type

    def print(self):
        """
        Support printing out the instruction description
        """
        if not self._is_two_qutrit_gate:
            if self.parameter is None:
                print("Gate " + str(self._type) + ", acting qutrit: " + str(self.first_qutrit))
            else:
                print("Gate " + str(self._type) + " with parameter " + str(self.parameter) +
                      ", acting qutrit: " + str(self.first_qutrit))
        else:
            print("Gate " + str(self._type) + ", acting qutrit: "
                  + str(self.first_qutrit) + ", control qutrit: " + str(self.second_qutrit))



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="qutritium",
    version="0.0.1",
    author="Son Pham, Tien Nguyen, Bao Bach",
    author_email="reachphamhson@gmail.com",
    description="Qutritium package provides basic calibration and features for qutrit system using Qiskit and other "
                "quantum packages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT License',
    url="https://github.com/spham1611/qutritium",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6"
)




# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Discriminator circuit and iq plot function """
from qiskit.circuit import Gate

from src.pulse import Pulse12
from src.pulse_creation import GateSchedule
from src.backend.backend_ibm import CustomProvider
from src.calibration.utility import _SetAttribute
from src.constant import QubitParameters

from typing import List


ghz_unit = QubitParameters.GHZ.value


class DiscriminatorQutrit(_SetAttribute):
    """ A Simple Discriminator class that set up circuit and plot the iq graph
    An example of using this class::

        from qutritium.calibration.discriminator import Discriminator
        from qutritium.backend.backend_ibm import EffProvider
        from qutritium.pulse import Pulse01, Pulse12

        efF_provider = EffProvider()
        pulse01 = Pulse01(duration=144, x_amp=0.2)
        pulse12 = Pulse12(pulse01=pulse01, duration=pulse01.duration, x_amp=pulse01.x_amp)

        ...Running Rough Rabi

        discriminator = Discriminator(eff_provider, pulse12, num_shots=4096)
        discriminator.prepare_circuit()
        discriminator.run_monitor()
        discriminator.plot_iq()

    Here is a list of available attributes of "DiscriminatorQutrit" class:
        * prepare_circuit(): prepares a discriminator with different QuantumCircuits
        * run_monitor(): run custom execute
        * plot_iq(): plot iq 012 graph

    To see more details, visit:
    ...
    """
    def __init__(self,
                 custom_provider: CustomProvider,
                 pulse_model: Pulse12,
                 backend_name: str = 'ibm_brisbane',
                 num_shots=4096,
                 delay_time: int = 22496) -> None:
        """ ctor

        Args:
            custom_provider: EffProvider instance which contains necessary information
            pulse_model: Pulse12 model only
            backend_name: name to choose which backend running on IBM server
            num_shots: number of experiments
            delay_time: sending a time delay pulse

        Notes:
            * model_space parameter is not necessary as we are discriminating states not processing
        """
        super().__init__(custom_provider=custom_provider,
                         pulse_model=pulse_model,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space='01')
        self.ramsey_frequency01 = self.pulse_model.pulse01.frequency
        self.ramsey_frequency12 = self.pulse_model.frequency
        self._delay_gate = Gate('5mus_delay', 1, [])
        self._x12 = Gate('X12', 1, [])

        # Schedules
        self._delay_schedule = GateSchedule.delay(backend=self.backend,
                                                  qubit=self.qubit,
                                                  delay_time=delay_time)

        self._x12_schedule = GateSchedule.x_amp_gaussian(backend=self.backend,
                                                         qubit=self.qubit,
                                                         pulse_model=self.pulse_model,
                                                         x_amp=self.pulse_model.x_amp)

        # Package
        self.package: List = []

    def prepare_circuit(self) -> None:
        """
        Creating a circuit that can represent a single qutrit state in IBM superconducting quantum computer.
        This helps us to visualize qutrit state with calibrated parameters.
        """
        from qiskit import QuantumCircuit

        ground_state_prep = QuantumCircuit(self.qubit + 1, self.cbit + 1)
        ground_state_prep.append(self._delay_gate, [self.qubit])
        ground_state_prep.add_calibration(self._delay_gate, [self.qubit], self._delay_schedule)
        # noinspection DuplicatedCode
        ground_state_prep.measure(self.qubit, self.cbit)

        first_excited_state_prep = QuantumCircuit(self.qubit + 1, self.cbit + 1)
        first_excited_state_prep.append(self._delay_gate, [self.qubit])
        first_excited_state_prep.add_calibration(self._delay_gate, [self.qubit], self._delay_schedule)
        first_excited_state_prep.x(self.qubit)
        # noinspection DuplicatedCode
        first_excited_state_prep.measure(self.qubit, self.cbit)

        second_excited_state_prep = QuantumCircuit(self.qubit + 1, self.cbit + 1)
        second_excited_state_prep.append(self._delay_gate, [self.qubit])
        second_excited_state_prep.add_calibration(self._delay_gate, [self.qubit], self._delay_schedule)
        second_excited_state_prep.x(self.qubit)
        second_excited_state_prep.append(self._x12, [self.qubit])
        second_excited_state_prep.add_calibration(self._x12, [self.qubit], self._x12_schedule)
        second_excited_state_prep.measure(self.qubit, self.cbit)

        self.package = [ground_state_prep, first_excited_state_prep, second_excited_state_prep]

    # noinspection DuplicatedCode
    def run_monitor(self,
                    num_shots: int = 4096,
                    meas_return='single',
                    meas_level=1,
                    **kwargs) -> None:
        """
        Run via run_monitor in abstract class
        Args:
            num_shots:
            meas_return:
            meas_level:
            **kwargs:

        Returns:

        """
        super().backend.run(num_shots=num_shots,
                            meas_return=meas_return,
                            meas_level=meas_level,
                            **kwargs)

    def plot_iq(self,
                x_min: int = -30,
                x_max: int = 30,
                y_min: int = -30,
                y_max: int = 30) -> None:
        """ Plot iq qutrit state

        Args:
            x_min: Min x-value of the plot
            x_max: Max x-value of the plot
            y_min: Min y-value of the plot
            y_max: Max y-value of the plot
        """
        import matplotlib.pyplot as plt
        from src.analyzer import DataAnalysis

        discriminator_job = self.backend.retrieve_job(self.submitted_job)
        discriminator_data = DataAnalysis(discriminator_job)
        discriminator_data.retrieve_data(True)
        discriminator_data.iq_012_plot(x_min, x_max, y_min, y_max)
        plt.savefig('output/0 1 2 Discrimination.svg', format='svg', dpi=1200)



# MIT License
#
# Copyright (c) [2023] [son pham, tien nguyen, bach bao]
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Rough rabi techniques"""
import numpy as np
from numpy.typing import NDArray

from qiskit.circuit import QuantumCircuit, Gate

from src.backend.backend_ibm import CustomProvider
from src.pulse import Pulse01, Pulse12
from src.calibration.utility import _SetAttribute
from src.exceptions.pulse_exception import MissingFrequencyPulse
from src.pulse_creation import GateSchedule

from typing import Optional, List, Union


class _RoughRabi(_SetAttribute):
    """ Provides basic attributes of RoughRabi technique and inherits SharedAttr available attributes
    A typical flow would be prepare_circuit -> run_monitor -> analyze IBMJob::

        from qutritium.calibration.rough_rabi import RoughRabi01, RoughRabi12

        pulse01 = Pulse01(duration=144, x_amp=0.2)
        pulse12 = Pulse12(pulse01=pulse01, duration=pulse01.duration, x_amp=pulse01.x_amp)

        ... (RoughRabi_children_class)

    Here is a list of available attributes of "_RoughRabi" class, excluding SharedAttr attr:
        * lambda_list: list of parameters for fit function
        * x_amp_sweeping_range: for sweeping amplitude
        * reset_sweeping_range():
        * prepare_circuit():
        * modify_pulse_model(): run analyze() and get the x_amp
        * save_data(): for saving data only
        * run_monitor():
        * analyze(): plots iq_data + get x_amp from fit function
    """

    def __init__(self,
                 pulse_model: Union[Pulse01, Pulse12],
                 custom_provider: CustomProvider,
                 backend_name: str,
                 num_shots: int,
                 model_space: str,
                 pulse_connected: Union[Pulse01, Pulse12]) -> None:
        """ _RoughRabi Constructor

        Notes:
            * This class should not be instantiated. Use RoughRabi01 or RoughRabi12

        Args:
            pulse_model:
            custom_provider:
            num_shots:
        """
        if pulse_model.frequency == 0:
            raise MissingFrequencyPulse('Need to have frequency. Assign it from TR function or default drive frequency')
        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space=model_space,
                         pulse_connected=pulse_connected)
        self.analyzer = None
        self._package: List = []

        # INTERNAL DESIGN ONLY
        self._lambda_list: Optional[List] = None
        self._x_amp_sweeping_range = np.linspace(-1., 1., 100)
        self._rabi_gate = Gate("Rabi", 1, [])
        self._rr_fit: Optional[NDArray] = None

    @property
    def lambda_list(self) -> List[float]:
        return self._lambda_list

    @lambda_list.setter
    def lambda_list(self, val_list: list) -> None:
        """

        Args:
            val_list:

        Returns:
            lambda_list:
        """
        if len(val_list) != 4:
            raise ValueError("Lambda list does not have sufficient elements")
        self._lambda_list = val_list

    @property
    def x_amp_sweeping_range(self) -> NDArray:
        return self._x_amp_sweeping_range

    @property
    def rr_fit(self) -> NDArray:
        return self._rr_fit

    def reset_sweeping_range(self, min_val: float, max_val: float, steps: int) -> None:
        """
        Reset the sweeping range
        Args:
            min_val:
            max_val:
            steps:
        """
        self._x_amp_sweeping_range = np.linspace(min_val, max_val, steps)

    def run_monitor(self,
                    num_shots: int = 0,
                    meas_level: int = 1,
                    meas_return: str = 'avg',
                    **kwargs) -> None:
        """
        Run custom execute()
        Args:
            num_shots:
            meas_level:
            meas_return:
            **kwargs:

        """
        from qiskit_ibm_provider.job import job_monitor
        from qiskit import execute

        self.num_shots = num_shots if num_shots != 0 else self.num_shots
        submitted_job = execute(experiments=self._package,
                                backend=self.backend,
                                meas_level=meas_level,
                                meas_return=meas_return,
                                shots=self.num_shots,
                                **kwargs)
        self.submitted_job = submitted_job.job_id()
        print(self.submitted_job)
        job_monitor(submitted_job)

    def analyze(self, job_id: str) -> float:
        """

        Args:
            job_id:

        Returns:

        """
        from src.utility import fit_function
        from src.analyzer import DataAnalysis

        if not job_id:
            experiment = self.custom_provider.retrieve_job(self.submitted_job)
        else:
            experiment = self.custom_provider.retrieve_job(job_id)
        self.analyzer = DataAnalysis(experiment)

        self.analyzer.retrieve_data(average=True)
        fit_params, self._rr_fit = fit_function(self._x_amp_sweeping_range, self.analyzer.IQ_data,
                                                lambda x, c1, c2, drive_period, phi:
                                                (c1 * np.cos(2 * np.pi * x / drive_period - phi) + c2),
                                                [5, 0, 0.5, 0])

        x_amp = (fit_params[2] / 2)
        return x_amp


class RoughRabi01(_RoughRabi):
    """ Provides rough rabi circuit and overrides methods from "_RoughRabi" class
    Here is an example of the rough rabi flow::

        from qutritium.calibration.rough_rabi import RoughRabi01
        from qutritium.backend.backend_ibm import EffProvider

        pulse01 = Pulse01(frequency=4.9e9, duration=144, x_amp=0.2)
        eff_provider = EffProvider()

        rr01 = RoughRabi01(pulse01, eff_provider)
        rr01.prepare_circuit()
        rr01.run_monitor()
        rr01.modify_pulse_model()

    Here is a list of available functions of "RoughRabi01" class, excluding inheritance attributes
        * prepare_circuit(): override "_RoughRabi"
        * modify_pulse_model(): override "_RoughRabi"

    Notes:
        * pulse_model should be processed by TR classes
    """

    def __init__(self,
                 pulse_model: Pulse01,
                 custom_provider: CustomProvider,
                 backend_name='ibm_brisbane',
                 num_shots: int = 4096) -> None:
        """
        Ctor
        Args:
            pulse_model:
            custom_provider:
            backend_name:
            num_shots:
        """
        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space='01',
                         pulse_connected=pulse_model.pulse12)
        self.lambda_list = [5, 0, 0.5, 0]

    def prepare_circuit(self) -> None:
        """ runs over x_amp sweeping range
        Returns:

        """
        self.pulse_model: Pulse01

        # Sweeping
        for x_amp in self.x_amp_sweeping_range:
            qc_rabi01 = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            # noinspection DuplicatedCode
            qc_rabi01.append(self._rabi_gate, [self.qubit])
            rabi_schedule = GateSchedule.x_amp_gaussian(
                backend=self.backend,
                x_amp=x_amp,
                pulse_model=self.pulse_model,
                qubit=self.qubit
            )
            qc_rabi01.measure(self.qubit, self.cbit)
            qc_rabi01.add_calibration(self._rabi_gate, (self.qubit,), rabi_schedule, x_amp)
            self._package.append(qc_rabi01)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """

        Args:
            job_id:

        Returns:

        """
        self.pulse_model: Pulse01
        self.pulse_model.x_amp = self.analyze(job_id=job_id)


class RoughRabi12(_RoughRabi):
    """ Overrides "_RoughRabi", used for Pulse12
    Here is an example of the rough rabi flow::

        from qutritium.calibration.rough_rabi import RoughRabi01
        from qutritium.backend.backend_ibm import EffProvider

        pulse01 = Pulse01(frequency=4.9e9, duration=144, x_amp=0.2)
        pulse12 = Pulse12(pulse01=pulse01, frequency=pulse01.frequency,
                          x_amp=pulse01.x_amp, duration=pulse01.duration)
        eff_provider = EffProvider()

        rr01 = RoughRabi01(pulse01, eff_provider)
        rr01.prepare_circuit()
        rr01.run_monitor()
        rr01.modify_pulse_model()

    Here is a list of available functions of "RoughRabi01" class, excluding inheritance attributes
        * prepare_circuit(): overrides "_RoughRabi"
        * modify_pulse_model(): overrides "_RoughRabi"
    """

    def __init__(self,
                 pulse_model: Pulse12,
                 custom_provider: CustomProvider,
                 backend_name='ibm_brisbane',
                 num_shots: int = 4096) -> None:
        """

        Args:
            pulse_model:
            custom_provider:
            backend_name:
            num_shots:
        """
        super().__init__(pulse_model=pulse_model,
                         custom_provider=custom_provider,
                         backend_name=backend_name,
                         num_shots=num_shots,
                         model_space='12',
                         pulse_connected=pulse_model.pulse01)
        self.lambda_list = [5, 0, 0.4, 0]

    def prepare_circuit(self) -> None:
        """ sweeps over x_amp with given range

        Returns:

        """
        self.pulse_model: Pulse12

        for x_amp in self.x_amp_sweeping_range:
            qc_rabi12 = QuantumCircuit(self.qubit + 1, self.cbit + 1)
            qc_rabi12.x(self.qubit)
            # noinspection DuplicatedCode
            qc_rabi12.append(self._rabi_gate, [self.qubit])
            rabi_schedule = GateSchedule.x_amp_gaussian(
                backend=self.backend,
                x_amp=x_amp,
                pulse_model=self.pulse_model,
                qubit=self.qubit
            )
            qc_rabi12.measure(self.qubit, self.cbit)
            qc_rabi12.add_calibration(self._rabi_gate, (self.qubit,), rabi_schedule, x_amp)
            self._package.append(qc_rabi12)

    def modify_pulse_model(self, job_id: str = None) -> None:
        """

        Args:
            job_id:

        Returns:

        """
        self.pulse_model: Pulse12
        self.pulse_model.x_amp = self.analyze(job_id=job_id)
